"""Public API for pivoteer."""

from __future__ import annotations

import logging
import zipfile
from pathlib import Path

import pandas as pd

from pivoteer.template_engine import TemplateEngine

LOGGER = logging.getLogger(__name__)


class Pivoteer:
    """Public entry point for applying DataFrames to Excel templates."""

    def __init__(
        self, template_path: str | Path, *, enable_pivot_field_sync: bool = False
    ) -> None:
        """Initialize with optional pivot cache field synchronization."""
        self._template_engine = TemplateEngine(Path(template_path))
        self._enable_pivot_field_sync = enable_pivot_field_sync

    def apply_dataframe(self, table_name: str, df: pd.DataFrame) -> None:
        """Apply a DataFrame to the specified table."""
        self._template_engine.apply_dataframe(table_name, df)

    def save(self, output_path: str | Path) -> Path:
        """Write the modified template to a new file."""
        output_path = Path(output_path)
        if self._enable_pivot_field_sync:
            self._template_engine.sync_pivot_cache_fields()
        self._template_engine.ensure_pivot_refresh_on_load()
        modified_parts = self._template_engine.get_modified_parts()

        with zipfile.ZipFile(self._template_engine.template_path, "r") as src:
            with zipfile.ZipFile(
                output_path, "w", compression=zipfile.ZIP_DEFLATED
            ) as dest:
                for info in src.infolist():
                    filename = info.filename
                    if filename in modified_parts:
                        dest.writestr(info, modified_parts[filename])
                    else:
                        dest.writestr(info, src.read(filename))

        LOGGER.info("Saved output to %s", output_path)
        return output_path



"""Generate a template, inject data, and save an updated report."""

from __future__ import annotations

import logging
from datetime import date, timedelta
from pathlib import Path
from typing import List

import pandas as pd

from pivoteer.core import Pivoteer
from tests.generate_dummy_template import generate_template


LOGGER = logging.getLogger(__name__)


def _build_dataframe(rows: int) -> pd.DataFrame:
    categories = ["Hardware", "Software", "Services"]
    regions = ["North", "South", "East", "West"]
    base = date(2024, 1, 1)

    data: List[dict[str, object]] = []
    for i in range(rows):
        data.append(
            {
                "Category": categories[i % len(categories)],
                "Region": regions[i % len(regions)],
                "Amount": 100.0 + (i * 10.0),
                "Date": base + timedelta(days=i),
            }
        )

    return pd.DataFrame(data)


def main() -> None:
    logging.basicConfig(level=logging.INFO, format="%(levelname)s %(message)s")

    template_path = Path("dummy_template.xlsx")
    output_path = Path("report_output.xlsx")

    generate_template(template_path)

    df = _build_dataframe(500)
    pivoteer = Pivoteer(template_path)
    pivoteer.apply_dataframe("DataSource", df)
    pivoteer.save(output_path)

    LOGGER.info("Report saved at %s", output_path)


if __name__ == "__main__":
    main()



"""Unit tests for pivot cache field synchronization."""

from __future__ import annotations

import zipfile
from pathlib import Path

import pandas as pd
import pytest
from lxml import etree

from pivoteer.core import Pivoteer
from pivoteer.exceptions import PivotCacheError
from pivoteer.pivot_cache_updater import sync_cache_fields
from pivoteer.utils import column_index_to_letter
from pivoteer.xml_engine import XmlEngine

_NS_MAIN = "http://schemas.openxmlformats.org/spreadsheetml/2006/main"
_NS_PKG_REL = "http://schemas.openxmlformats.org/package/2006/relationships"


def _write_minimal_xlsx(
    path: Path, table_columns: list[str], cache_fields: list[str]
) -> None:
    _write_minimal_xlsx_with_caches(
        path,
        table_columns=table_columns,
        cache_definitions=[
            {
                "path": "xl/pivotCache/pivotCacheDefinition1.xml",
                "table_name": "DataSource",
                "cache_fields": cache_fields,
            }
        ],
    )


def _write_minimal_xlsx_with_caches(
    path: Path,
    *,
    table_columns: list[str],
    cache_definitions: list[dict],
    include_table_columns: bool = True,
) -> None:
    ref = f"A1:{column_index_to_letter(len(table_columns))}2"
    table_columns_xml = "".join(
        f'<tableColumn id="{idx}" name="{name}"/>'
        for idx, name in enumerate(table_columns, start=1)
    )

    workbook_xml = (
        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
        f'<workbook xmlns="{_NS_MAIN}" '
        'xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships">'
        "<sheets>"
        '<sheet name="Data" sheetId="1" r:id="rId1"/>'
        "</sheets>"
        "</workbook>"
    )

    pivot_rels = "".join(
        f'<Relationship Id="rId{idx}" '
        'Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/pivotCacheDefinition" '
        f'Target="{cache_def["path"].removeprefix("xl/")}" />'
        for idx, cache_def in enumerate(cache_definitions, start=2)
    )
    workbook_rels_xml = "".join(
        [
            '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>',
            f'<Relationships xmlns="{_NS_PKG_REL}">',
            '<Relationship Id="rId1" ',
            'Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/worksheet" ',
            'Target="worksheets/sheet1.xml"/>',
            pivot_rels,
            "</Relationships>",
        ]
    )

    worksheet_xml = (
        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
        f'<worksheet xmlns="{_NS_MAIN}" '
        'xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships">'
        "<sheetData/>"
        '<tableParts count="1">'
        '<tablePart r:id="rId1"/>'
        "</tableParts>"
        "</worksheet>"
    )

    worksheet_rels_xml = (
        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
        f'<Relationships xmlns="{_NS_PKG_REL}">'
        '<Relationship Id="rId1" '
        'Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/table" '
        'Target="../tables/table1.xml"/>'
        "</Relationships>"
    )

    table_columns_node = ""
    if include_table_columns:
        table_columns_node = "".join(
            [
                f'<tableColumns count="{len(table_columns)}">',
                table_columns_xml,
                "</tableColumns>",
            ]
        )
    table_xml = "".join(
        [
            '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>',
            f'<table xmlns="{_NS_MAIN}" id="1" name="DataSource" ',
            f'displayName="DataSource" ref="{ref}" totalsRowShown="0">',
            f'<autoFilter ref="{ref}"/>',
            table_columns_node,
            "</table>",
        ]
    )

    with zipfile.ZipFile(path, "w", compression=zipfile.ZIP_DEFLATED) as archive:
        archive.writestr("xl/workbook.xml", workbook_xml)
        archive.writestr("xl/_rels/workbook.xml.rels", workbook_rels_xml)
        archive.writestr("xl/worksheets/sheet1.xml", worksheet_xml)
        archive.writestr("xl/worksheets/_rels/sheet1.xml.rels", worksheet_rels_xml)
        archive.writestr("xl/tables/table1.xml", table_xml)
        for cache_def in cache_definitions:
            cache_fields_xml = "".join(
                f'<cacheField name="{name}"/>'
                for name in cache_def.get("cache_fields", [])
            )
            cache_fields_node = ""
            if cache_def.get("include_cache_fields", True):
                cache_fields_node = (
                    f'<cacheFields count="{len(cache_def.get("cache_fields", []))}">'
                    f"{cache_fields_xml}"
                    "</cacheFields>"
                )
            pivot_cache_xml = (
                '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                f'<pivotCacheDefinition xmlns="{_NS_MAIN}" refreshOnLoad="1">'
                '<cacheSource type="worksheet">'
                f'<worksheetSource name="{cache_def["table_name"]}" '
                f'sheet="Data" ref="{ref}"/>'
                "</cacheSource>"
                f"{cache_fields_node}"
                "</pivotCacheDefinition>"
            )
            archive.writestr(cache_def["path"], pivot_cache_xml)


def _main_namespace(tree: etree._ElementTree) -> str:
    root = tree.getroot()
    return root.nsmap.get(None) or root.nsmap.get("main") or _NS_MAIN


def _cache_fields(tree: etree._ElementTree) -> list[etree._Element]:
    ns = _main_namespace(tree)
    cache_fields = tree.find(f".//{{{ns}}}cacheFields")
    if cache_fields is None:
        return []
    return cache_fields.findall(f"{{{ns}}}cacheField")


def _cache_field_names(tree: etree._ElementTree) -> list[str]:
    return [node.get("name") for node in _cache_fields(tree) if node.get("name")]


def _read_xml(archive: zipfile.ZipFile, path: str) -> etree._ElementTree:
    parser = etree.XMLParser(remove_blank_text=False)
    return etree.fromstring(archive.read(path), parser).getroottree()


def test_cache_fields_unchanged_when_schema_matches(tmp_path: Path) -> None:
    fixture = tmp_path / "baseline.xlsx"
    columns = ["Category", "Region", "Amount"]
    _write_minimal_xlsx(fixture, columns, columns)

    workbook_map = XmlEngine(fixture).build_workbook_map()
    updates = sync_cache_fields(workbook_map, "DataSource")

    assert updates == {}


def test_cache_fields_appended_for_new_columns(tmp_path: Path) -> None:
    fixture = tmp_path / "new_column.xlsx"
    table_columns = ["Category", "Region", "Amount", "call_type"]
    cache_fields = ["Category", "Region", "Amount"]
    _write_minimal_xlsx(fixture, table_columns, cache_fields)

    workbook_map = XmlEngine(fixture).build_workbook_map()
    updates = sync_cache_fields(workbook_map, "DataSource")

    assert "xl/pivotCache/pivotCacheDefinition1.xml" in updates
    tree = updates["xl/pivotCache/pivotCacheDefinition1.xml"]
    cache_field_elements = _cache_fields(tree)
    assert _cache_field_names(tree) == table_columns
    assert len(cache_field_elements) == len(table_columns)
    ns = _main_namespace(tree)
    cache_fields_node = tree.find(f".//{{{ns}}}cacheFields")
    assert cache_fields_node is not None
    assert cache_fields_node.get("count") == str(len(table_columns))


def test_cache_fields_append_preserves_existing_order(tmp_path: Path) -> None:
    fixture = tmp_path / "order_preserved.xlsx"
    table_columns = ["Category", "Region", "Amount", "call_type"]
    cache_fields = ["Region", "Amount", "Extra"]
    _write_minimal_xlsx(fixture, table_columns, cache_fields)

    workbook_map = XmlEngine(fixture).build_workbook_map()
    updates = sync_cache_fields(workbook_map, "DataSource")

    tree = updates["xl/pivotCache/pivotCacheDefinition1.xml"]
    assert _cache_field_names(tree) == [
        "Region",
        "Amount",
        "Extra",
        "Category",
        "call_type",
    ]


def test_cache_fields_only_updates_matching_table(tmp_path: Path) -> None:
    fixture = tmp_path / "multiple_caches.xlsx"
    table_columns = ["Category", "Region", "Amount", "call_type"]
    cache_definitions = [
        {
            "path": "xl/pivotCache/pivotCacheDefinition1.xml",
            "table_name": "DataSource",
            "cache_fields": ["Category", "Region", "Amount"],
        },
        {
            "path": "xl/pivotCache/pivotCacheDefinition2.xml",
            "table_name": "OtherTable",
            "cache_fields": ["Foo", "Bar"],
        },
    ]
    _write_minimal_xlsx_with_caches(
        fixture, table_columns=table_columns, cache_definitions=cache_definitions
    )

    workbook_map = XmlEngine(fixture).build_workbook_map()
    updates = sync_cache_fields(workbook_map, "DataSource")

    assert list(updates) == ["xl/pivotCache/pivotCacheDefinition1.xml"]
    tree = updates["xl/pivotCache/pivotCacheDefinition1.xml"]
    assert _cache_field_names(tree) == table_columns


def test_missing_table_columns_raises(tmp_path: Path) -> None:
    fixture = tmp_path / "missing_table_columns.xlsx"
    cache_definitions = [
        {
            "path": "xl/pivotCache/pivotCacheDefinition1.xml",
            "table_name": "DataSource",
            "cache_fields": ["Category"],
        }
    ]
    _write_minimal_xlsx_with_caches(
        fixture,
        table_columns=["Category"],
        cache_definitions=cache_definitions,
        include_table_columns=False,
    )

    workbook_map = XmlEngine(fixture).build_workbook_map()
    with pytest.raises(PivotCacheError):
        sync_cache_fields(workbook_map, "DataSource")


def test_missing_cache_fields_raises(tmp_path: Path) -> None:
    fixture = tmp_path / "missing_cache_fields.xlsx"
    cache_definitions = [
        {
            "path": "xl/pivotCache/pivotCacheDefinition1.xml",
            "table_name": "DataSource",
            "cache_fields": ["Category"],
            "include_cache_fields": False,
        }
    ]
    _write_minimal_xlsx_with_caches(
        fixture,
        table_columns=["Category"],
        cache_definitions=cache_definitions,
    )

    workbook_map = XmlEngine(fixture).build_workbook_map()
    with pytest.raises(PivotCacheError):
        sync_cache_fields(workbook_map, "DataSource")


def test_opt_in_flag_leaves_cache_fields_unchanged(tmp_path: Path) -> None:
    fixture = tmp_path / "opt_out.xlsx"
    table_columns = ["Category", "Region", "Amount", "call_type"]
    cache_fields = ["Category", "Region", "Amount"]
    _write_minimal_xlsx(fixture, table_columns, cache_fields)

    df = pd.DataFrame(
        [
            {
                "Category": "Hardware",
                "Region": "North",
                "Amount": 100.0,
                "call_type": "external",
            }
        ]
    )
    output_path = tmp_path / "opt_out_output.xlsx"
    pivoteer = Pivoteer(fixture, enable_pivot_field_sync=False)
    pivoteer.apply_dataframe("DataSource", df)
    pivoteer.save(output_path)

    with zipfile.ZipFile(output_path, "r") as archive:
        tree = _read_xml(archive, "xl/pivotCache/pivotCacheDefinition1.xml")
        assert _cache_field_names(tree) == cache_fields



"""Pytest fixtures for pivoteer tests."""

from __future__ import annotations

import sys
from pathlib import Path

import pytest

PROJECT_ROOT = Path(__file__).resolve().parents[1]
SRC_ROOT = PROJECT_ROOT / "src"
if str(SRC_ROOT) not in sys.path:
    sys.path.insert(0, str(SRC_ROOT))


@pytest.fixture()
def template_path(tmp_path: Path) -> Path:
    from tests.generate_dummy_template import generate_template

    template = tmp_path / "dummy_template.xlsx"
    generate_template(template)
    return template



"""Utility helpers for Excel XML manipulation."""

from __future__ import annotations

import re

_A1_RE = re.compile(r"^([A-Z]+)([0-9]+)$")
_A1_RANGE_RE = re.compile(r"^([A-Z]+[0-9]+):([A-Z]+[0-9]+)$")


def column_index_to_letter(index: int) -> str:
    """Convert a 1-based column index to Excel column letters."""
    if index < 1:
        raise ValueError("Column index must be >= 1.")

    letters = []
    current = index
    while current:
        current, remainder = divmod(current - 1, 26)
        letters.append(chr(65 + remainder))
    return "".join(reversed(letters))


def column_letter_to_index(letters: str) -> int:
    """Convert Excel column letters to a 1-based column index."""
    if not letters or not letters.isalpha() or not letters.isupper():
        raise ValueError(f"Invalid column letters: {letters!r}")

    index = 0
    for char in letters:
        index = index * 26 + (ord(char) - 64)
    return index


def parse_a1_cell(cell: str) -> tuple[int, int]:
    """Parse an A1 cell reference into (row, col) 1-based coordinates."""
    match = _A1_RE.match(cell)
    if not match:
        raise ValueError(f"Invalid A1 cell reference: {cell!r}")

    col_letters, row_str = match.groups()
    return int(row_str), column_letter_to_index(col_letters)


def parse_a1_range(a1_range: str) -> tuple[tuple[int, int], tuple[int, int]]:
    """Parse an A1 range like A1:D10 into row/col tuples."""
    match = _A1_RANGE_RE.match(a1_range)
    if not match:
        raise ValueError(f"Invalid A1 range: {a1_range!r}")

    start_cell, end_cell = match.groups()
    return parse_a1_cell(start_cell), parse_a1_cell(end_cell)


def build_a1_cell(row: int, col: int) -> str:
    """Build an A1 cell reference from 1-based row/col."""
    if row < 1 or col < 1:
        raise ValueError("Row and column must be >= 1.")
    return f"{column_index_to_letter(col)}{row}"


def build_a1_range(start_row: int, start_col: int, end_row: int, end_col: int) -> str:
    """Build an A1 range from 1-based coordinates."""
    if end_row < start_row or end_col < start_col:
        raise ValueError("Range end must be >= range start.")
    return f"{build_a1_cell(start_row, start_col)}:{build_a1_cell(end_row, end_col)}"



"""Generate a dummy Excel template with a table, pivot table, and slicer."""

from __future__ import annotations

import argparse
import logging
from datetime import date, timedelta
from pathlib import Path

import xlsxwriter

LOGGER = logging.getLogger(__name__)


def _seed_rows() -> list[list[object]]:
    categories = ["Hardware", "Software", "Services"]
    regions = ["North", "South", "East", "West"]
    base = date(2024, 1, 1)
    rows: list[list[object]] = []
    amount = 100.0

    for i in range(12):
        rows.append(
            [
                categories[i % len(categories)],
                regions[i % len(regions)],
                amount + (i * 25.0),
                base + timedelta(days=i),
            ]
        )
    return rows


def _write_table(
    worksheet: xlsxwriter.worksheet.Worksheet,
    headers: list[str],
    rows: list[list[object]],
) -> str:
    worksheet.write_row(0, 0, headers)
    for row_idx, row in enumerate(rows, start=1):
        worksheet.write_row(row_idx, 0, row)

    last_row = len(rows)
    last_col = len(headers) - 1

    table_range = xlsxwriter.utility.xl_range(0, 0, last_row, last_col)
    worksheet.add_table(
        table_range,
        {
            "name": "DataSource",
            "columns": [{"header": h} for h in headers],
        },
    )
    return table_range


def _add_pivot_table(
    worksheet: xlsxwriter.worksheet.Worksheet, data_range: str
) -> str | None:
    pivot_table_name = "PivotTable1"
    if hasattr(worksheet, "add_pivot_table"):
        worksheet.add_pivot_table(
            {
                "data": data_range,
                "name": pivot_table_name,
                "rows": [{"data": "Category"}],
                "columns": [{"data": "Region"}],
                "values": [{"data": "Amount", "subtotal": "sum"}],
            }
        )
        return pivot_table_name

    LOGGER.warning("Pivot table API not available in installed xlsxwriter.")
    return None


def _add_slicer(
    worksheet: xlsxwriter.worksheet.Worksheet,
    pivot_table_name: str | None,
    field_name: str,
) -> None:
    if not pivot_table_name:
        return
    options = {"pivot_table": pivot_table_name, "field": field_name}

    if hasattr(worksheet, "insert_slicer"):
        try:
            worksheet.insert_slicer("H2", options)
            return
        except TypeError:
            pass

    if hasattr(worksheet, "add_slicer"):
        try:
            worksheet.add_slicer(options)
            return
        except TypeError:
            pass

    LOGGER.warning("Slicer API not available in installed xlsxwriter.")


def generate_template(output_path: Path) -> None:
    headers = ["Category", "Region", "Amount", "Date"]
    rows = _seed_rows()

    with xlsxwriter.Workbook(output_path) as workbook:
        data_sheet = workbook.add_worksheet("Data")
        pivot_sheet = workbook.add_worksheet("Pivot")

        data_range = _write_table(data_sheet, headers, rows)
        pivot_table_name = _add_pivot_table(pivot_sheet, data_range)
        _add_slicer(pivot_sheet, pivot_table_name, "Region")

    LOGGER.info("Template generated at %s", output_path)


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Generate a dummy Excel template with table, pivot, slicer."
    )
    parser.add_argument(
        "--output",
        type=Path,
        default=Path("dummy_template.xlsx"),
        help="Output .xlsx path",
    )
    return parser.parse_args()


def main() -> None:
    logging.basicConfig(level=logging.INFO, format="%(levelname)s %(message)s")
    args = _parse_args()
    generate_template(args.output)


if __name__ == "__main__":
    main()



"""Unit tests for utility helpers."""

from __future__ import annotations

import pytest

from pivoteer.utils import (
    build_a1_cell,
    build_a1_range,
    column_index_to_letter,
    column_letter_to_index,
    parse_a1_cell,
    parse_a1_range,
)


class TestColumnIndexToLetter:
    def test_single_letters(self) -> None:
        assert column_index_to_letter(1) == "A"
        assert column_index_to_letter(26) == "Z"

    def test_double_letters(self) -> None:
        assert column_index_to_letter(27) == "AA"
        assert column_index_to_letter(28) == "AB"
        assert column_index_to_letter(52) == "AZ"
        assert column_index_to_letter(53) == "BA"
        assert column_index_to_letter(702) == "ZZ"

    def test_triple_letters(self) -> None:
        assert column_index_to_letter(703) == "AAA"

    def test_excel_max_column(self) -> None:
        assert column_index_to_letter(16384) == "XFD"

    def test_zero_raises(self) -> None:
        with pytest.raises(ValueError):
            column_index_to_letter(0)

    def test_negative_raises(self) -> None:
        with pytest.raises(ValueError):
            column_index_to_letter(-1)


class TestColumnLetterToIndex:
    def test_single_letters(self) -> None:
        assert column_letter_to_index("A") == 1
        assert column_letter_to_index("Z") == 26

    def test_double_letters(self) -> None:
        assert column_letter_to_index("AA") == 27
        assert column_letter_to_index("AZ") == 52
        assert column_letter_to_index("ZZ") == 702

    def test_triple_letters(self) -> None:
        assert column_letter_to_index("AAA") == 703
        assert column_letter_to_index("XFD") == 16384

    def test_lowercase_raises(self) -> None:
        with pytest.raises(ValueError):
            column_letter_to_index("a")

    def test_empty_raises(self) -> None:
        with pytest.raises(ValueError):
            column_letter_to_index("")

    def test_digits_raises(self) -> None:
        with pytest.raises(ValueError):
            column_letter_to_index("A1")


class TestRoundTrip:
    @pytest.mark.parametrize("index", [1, 13, 26, 27, 52, 256, 702, 703, 16384])
    def test_column_round_trip(self, index: int) -> None:
        letters = column_index_to_letter(index)
        assert column_letter_to_index(letters) == index


class TestParseA1Cell:
    def test_simple(self) -> None:
        assert parse_a1_cell("A1") == (1, 1)

    def test_double_letter(self) -> None:
        assert parse_a1_cell("AA100") == (100, 27)

    def test_invalid_raises(self) -> None:
        with pytest.raises(ValueError):
            parse_a1_cell("1A")

    def test_empty_raises(self) -> None:
        with pytest.raises(ValueError):
            parse_a1_cell("")


class TestBuildA1Cell:
    def test_simple(self) -> None:
        assert build_a1_cell(1, 1) == "A1"
        assert build_a1_cell(10, 3) == "C10"

    def test_zero_row_raises(self) -> None:
        with pytest.raises(ValueError):
            build_a1_cell(0, 1)

    def test_zero_col_raises(self) -> None:
        with pytest.raises(ValueError):
            build_a1_cell(1, 0)


class TestParseA1Range:
    def test_simple(self) -> None:
        result = parse_a1_range("A1:D10")
        assert result == ((1, 1), (10, 4))

    def test_invalid_raises(self) -> None:
        with pytest.raises(ValueError):
            parse_a1_range("A1")

    def test_single_cell_range(self) -> None:
        result = parse_a1_range("B2:B2")
        assert result == ((2, 2), (2, 2))


class TestBuildA1Range:
    def test_simple(self) -> None:
        assert build_a1_range(1, 1, 10, 4) == "A1:D10"

    def test_end_before_start_raises(self) -> None:
        with pytest.raises(ValueError):
            build_a1_range(10, 1, 5, 4)

    def test_round_trip(self) -> None:
        original = "B3:F20"
        parsed = parse_a1_range(original)
        rebuilt = build_a1_range(parsed[0][0], parsed[0][1], parsed[1][0], parsed[1][1])
        assert rebuilt == original



"""Unit tests for XmlEngine cell value injection and row handling."""

from __future__ import annotations

from datetime import date, datetime
from math import nan

import pandas as pd
import pytest
from lxml import etree

from pivoteer.exceptions import InvalidDataError, XmlStructureError
from pivoteer.xml_engine import XmlEngine, read_xml_part

_NS_MAIN = "http://schemas.openxmlformats.org/spreadsheetml/2006/main"
_NSMAP = {"main": _NS_MAIN}


def _make_sheet_tree() -> etree._ElementTree:
    """Create a minimal worksheet tree with an empty sheetData."""
    xml = f'<worksheet xmlns="{_NS_MAIN}"><sheetData/></worksheet>'
    return etree.fromstring(xml.encode()).getroottree()


def _get_cell(tree: etree._ElementTree, cell_ref: str) -> etree._Element | None:
    return tree.find(f".//main:row/main:c[@r='{cell_ref}']", namespaces=_NSMAP)


class TestInjectRowsInlineStrings:
    def _make_engine(self, tmp_path) -> XmlEngine:
        # Create a minimal xlsx for XmlEngine init
        import zipfile

        path = tmp_path / "minimal.xlsx"
        workbook_xml = (
            f'<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
            f'<workbook xmlns="{_NS_MAIN}" '
            'xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships">'
            "<sheets>"
            '<sheet name="Data" sheetId="1" r:id="rId1"/>'
            "</sheets>"
            "</workbook>"
        )
        rels_xml = (
            '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
            '<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships">'
            '<Relationship Id="rId1" '
            'Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/worksheet" '
            'Target="worksheets/sheet1.xml"/>'
            "</Relationships>"
        )
        worksheet_xml = (
            f'<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
            f'<worksheet xmlns="{_NS_MAIN}"><sheetData/></worksheet>'
        )
        with zipfile.ZipFile(path, "w") as archive:
            archive.writestr("xl/workbook.xml", workbook_xml)
            archive.writestr("xl/_rels/workbook.xml.rels", rels_xml)
            archive.writestr("xl/worksheets/sheet1.xml", worksheet_xml)
        return XmlEngine(path)

    def test_integer_value(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        engine.inject_rows_inline_strings(tree, 1, 1, [[42]])
        cell = _get_cell(tree, "A1")
        assert cell is not None
        assert cell.get("t") is None  # numeric, no type attr
        v = cell.find(f"{{{_NS_MAIN}}}v")
        assert v is not None
        assert v.text == "42"

    def test_float_value(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        engine.inject_rows_inline_strings(tree, 1, 1, [[3.14]])
        cell = _get_cell(tree, "A1")
        assert cell is not None
        v = cell.find(f"{{{_NS_MAIN}}}v")
        assert v is not None
        assert v.text == "3.14"

    def test_string_value(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        engine.inject_rows_inline_strings(tree, 1, 1, [["hello"]])
        cell = _get_cell(tree, "A1")
        assert cell is not None
        assert cell.get("t") == "inlineStr"
        t = cell.find(f".//{{{_NS_MAIN}}}t")
        assert t is not None
        assert t.text == "hello"

    def test_none_value(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        engine.inject_rows_inline_strings(tree, 1, 1, [[None]])
        cell = _get_cell(tree, "A1")
        assert cell is not None
        assert cell.get("t") is None
        assert len(list(cell)) == 0  # no children

    def test_nan_value(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        engine.inject_rows_inline_strings(tree, 1, 1, [[nan]])
        cell = _get_cell(tree, "A1")
        assert cell is not None
        assert cell.get("t") is None
        assert len(list(cell)) == 0  # treated as missing

    def test_pd_nat_value(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        engine.inject_rows_inline_strings(tree, 1, 1, [[pd.NaT]])
        cell = _get_cell(tree, "A1")
        assert cell is not None
        assert len(list(cell)) == 0

    def test_date_value(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        engine.inject_rows_inline_strings(tree, 1, 1, [[date(2024, 6, 15)]])
        cell = _get_cell(tree, "A1")
        assert cell is not None
        assert cell.get("t") == "inlineStr"
        t = cell.find(f".//{{{_NS_MAIN}}}t")
        assert t is not None
        assert t.text == "2024-06-15"

    def test_datetime_value(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        engine.inject_rows_inline_strings(tree, 1, 1, [[datetime(2024, 6, 15, 10, 30)]])
        cell = _get_cell(tree, "A1")
        assert cell is not None
        t = cell.find(f".//{{{_NS_MAIN}}}t")
        assert t is not None
        assert "2024-06-15" in t.text

    def test_multiple_columns(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        engine.inject_rows_inline_strings(tree, 1, 1, [[1, "two", 3.0]])
        assert _get_cell(tree, "A1") is not None
        assert _get_cell(tree, "B1") is not None
        assert _get_cell(tree, "C1") is not None

    def test_multiple_rows(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        engine.inject_rows_inline_strings(tree, 2, 1, [["a"], ["b"], ["c"]])
        assert _get_cell(tree, "A2") is not None
        assert _get_cell(tree, "A3") is not None
        assert _get_cell(tree, "A4") is not None

    def test_zero_start_row_raises(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        with pytest.raises(InvalidDataError):
            engine.inject_rows_inline_strings(tree, 0, 1, [["x"]])

    def test_empty_rows_no_change(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        engine.inject_rows_inline_strings(tree, 1, 1, [])
        rows = tree.findall(".//main:row", namespaces=_NSMAP)
        assert len(rows) == 0

    def test_rows_are_sorted(self, tmp_path) -> None:
        engine = self._make_engine(tmp_path)
        tree = _make_sheet_tree()
        # Inject at row 5 then row 2 to test sorting
        engine.inject_rows_inline_strings(tree, 5, 1, [["late"]])
        engine.inject_rows_inline_strings(tree, 2, 1, [["early"]])
        rows = tree.findall(".//main:row", namespaces=_NSMAP)
        row_indices = [int(r.get("r", "0")) for r in rows]
        assert row_indices == sorted(row_indices)


class TestReadXmlPart:
    def test_missing_path_raises(self, tmp_path) -> None:
        import zipfile

        path = tmp_path / "empty.zip"
        with zipfile.ZipFile(path, "w"):
            pass

        with zipfile.ZipFile(path, "r") as archive:
            with pytest.raises(XmlStructureError, match="Missing XML part"):
                read_xml_part(archive, "nonexistent.xml")

    def test_reads_valid_xml(self, tmp_path) -> None:
        import zipfile

        path = tmp_path / "test.zip"
        xml_content = '<?xml version="1.0"?><root><child/></root>'
        with zipfile.ZipFile(path, "w") as archive:
            archive.writestr("test.xml", xml_content)

        with zipfile.ZipFile(path, "r") as archive:
            tree = read_xml_part(archive, "test.xml")
            assert tree.getroot().tag == "root"



"""Integration test to verify resized table range in output XLSX."""

from __future__ import annotations

import posixpath
import zipfile
from pathlib import Path

from lxml import etree

from pivoteer.core import Pivoteer
from pivoteer.utils import parse_a1_range

_NS_MAIN = "http://schemas.openxmlformats.org/spreadsheetml/2006/main"
_NS_REL = "http://schemas.openxmlformats.org/officeDocument/2006/relationships"
_NS_PKG_REL = "http://schemas.openxmlformats.org/package/2006/relationships"

_NSMAP_MAIN = {"main": _NS_MAIN}
_NSMAP_PKG = {"rel": _NS_PKG_REL}


def _read_xml(archive: zipfile.ZipFile, path: str) -> etree._ElementTree:
    data = archive.read(path)
    parser = etree.XMLParser(remove_blank_text=False)
    return etree.fromstring(data, parser).getroottree()


def _parse_relationships(rels_tree: etree._ElementTree) -> dict[str, str]:
    rels: dict[str, str] = {}
    rel_nodes = rels_tree.findall(".//rel:Relationship", namespaces=_NSMAP_PKG)
    for rel in rel_nodes:
        rel_id = rel.get("Id")
        target = rel.get("Target")
        if rel_id and target:
            rels[rel_id] = target
    return rels


def _resolve_table_path(archive: zipfile.ZipFile, table_name: str) -> str:
    workbook_tree = _read_xml(archive, "xl/workbook.xml")
    rels_tree = _read_xml(archive, "xl/_rels/workbook.xml.rels")
    rel_map = _parse_relationships(rels_tree)

    sheet_nodes = workbook_tree.findall(".//main:sheets/main:sheet", _NSMAP_MAIN)
    for sheet in sheet_nodes:
        rel_id = sheet.get(f"{{{_NS_REL}}}id")
        if not rel_id:
            continue
        worksheet_target = rel_map.get(rel_id)
        if not worksheet_target:
            continue

        worksheet_path = f"xl/{worksheet_target}"
        rels_path = (
            Path(worksheet_path).parent / "_rels" / f"{Path(worksheet_path).name}.rels"
        )
        rels_tree = _read_xml(archive, str(rels_path))
        worksheet_rel_map = _parse_relationships(rels_tree)

        sheet_tree = _read_xml(archive, worksheet_path)
        table_parts = sheet_tree.findall(
            ".//main:tableParts/main:tablePart", _NSMAP_MAIN
        )
        for table_part in table_parts:
            table_rel_id = table_part.get(f"{{{_NS_REL}}}id")
            if not table_rel_id:
                continue
            target = worksheet_rel_map.get(table_rel_id)
            if not target:
                continue

            base_dir = posixpath.dirname(worksheet_path)
            table_path = posixpath.normpath(posixpath.join(base_dir, target))
            if not table_path.startswith("xl/"):
                table_path = f"xl/{table_path.lstrip('./')}"

            table_tree = _read_xml(archive, table_path)
            table_node = table_tree.getroot()
            name = table_node.get("name")
            if name == table_name:
                return table_path

    raise RuntimeError(f"Table {table_name!r} not found.")


def _build_output_report(template_path: Path, output_path: Path) -> None:
    from datetime import date, timedelta

    import pandas as pd

    data = []
    base = date(2024, 1, 1)
    for i in range(500):
        data.append(
            {
                "Category": "Hardware",
                "Region": "North",
                "Amount": 100.0 + (i * 10.0),
                "Date": base + timedelta(days=i),
            }
        )

    df = pd.DataFrame(data)
    pivoteer = Pivoteer(template_path)
    pivoteer.apply_dataframe("DataSource", df)
    pivoteer.save(output_path)


def test_table_range_resized(template_path: Path, tmp_path: Path) -> None:
    report_path = tmp_path / "report_output.xlsx"
    _build_output_report(template_path, report_path)

    with zipfile.ZipFile(report_path, "r") as archive:
        table_path = _resolve_table_path(archive, "DataSource")
        table_tree = _read_xml(archive, table_path)
        table_node = table_tree.getroot()
        ref = table_node.get("ref")
        if not ref:
            raise RuntimeError("Table ref attribute missing.")

    (start_row, _), (end_row, _) = parse_a1_range(ref)
    row_count = end_row - start_row + 1

    assert row_count >= 501, (
        f"Expected at least 501 rows (header + 500 data). Got {row_count}."
    )


def test_pivot_refresh_flag_optional(template_path: Path, tmp_path: Path) -> None:
    report_path = tmp_path / "report_output.xlsx"
    _build_output_report(template_path, report_path)

    pivot_refresh_enabled = False
    pivot_cache_found = False
    with zipfile.ZipFile(report_path, "r") as archive:
        for filename in archive.namelist():
            if filename.startswith(
                "xl/pivotCache/pivotCacheDefinition"
            ) and filename.endswith(".xml"):
                pivot_cache_found = True
                pivot_tree = _read_xml(archive, filename)
                root = pivot_tree.getroot()
                if root.get("refreshOnLoad") == "1":
                    pivot_refresh_enabled = True
                    break

    if pivot_cache_found:
        assert pivot_refresh_enabled, "Pivot cache refreshOnLoad is not enabled."



"""Pivot cache field synchronization for Excel tables."""

from __future__ import annotations

import zipfile

from lxml import etree

from pivoteer.exceptions import PivotCacheError, TableNotFoundError
from pivoteer.models import WorkbookMap
from pivoteer.xml_engine import read_xml_part

_NS_MAIN = "http://schemas.openxmlformats.org/spreadsheetml/2006/main"


def sync_cache_fields(
    workbook_map: WorkbookMap, table_name: str
) -> dict[str, etree._ElementTree]:
    """Sync pivot cache field names with the specified table's columns.

    Returns a mapping of modified pivot cache definition paths to XML trees.
    """
    table_ref = workbook_map.tables.get(table_name)
    if not table_ref:
        raise TableNotFoundError(f"Table not found: {table_name}")

    cache_paths = list(workbook_map.pivot_cache_definition_paths.values())
    if not cache_paths:
        return {}

    with zipfile.ZipFile(workbook_map.template_path, "r") as archive:
        table_tree = read_xml_part(archive, table_ref.table_path)
        table_columns = _extract_table_columns(table_tree)

        updated_parts: dict[str, etree._ElementTree] = {}
        for cache_path in cache_paths:
            cache_tree = read_xml_part(archive, cache_path)
            if _cache_source_table_name(cache_tree) != table_name:
                continue
            updated = _append_missing_cache_fields(cache_tree, table_columns)
            if updated:
                updated_parts[cache_path] = cache_tree

        return updated_parts


def _extract_table_columns(table_tree: etree._ElementTree) -> list[str]:
    ns = _get_main_namespace(table_tree)
    table_columns = table_tree.find(f".//{{{ns}}}tableColumns")
    if table_columns is None:
        raise PivotCacheError("Table columns element not found.")

    columns: list[str] = []
    for column in table_columns.findall(f"{{{ns}}}tableColumn"):
        name = column.get("name")
        if name:
            columns.append(name)

    if not columns:
        msg = "Table columns are empty; cannot sync pivot cache fields."
        raise PivotCacheError(msg)
    return columns


def _cache_source_table_name(cache_tree: etree._ElementTree) -> str | None:
    ns = _get_main_namespace(cache_tree)
    worksheet_source = cache_tree.find(
        f".//{{{ns}}}cacheSource/{{{ns}}}worksheetSource"
    )
    if worksheet_source is None:
        return None
    return worksheet_source.get("name")


def _append_missing_cache_fields(
    cache_tree: etree._ElementTree, table_columns: list[str]
) -> bool:
    ns = _get_main_namespace(cache_tree)
    cache_fields = cache_tree.find(f".//{{{ns}}}cacheFields")
    if cache_fields is None:
        msg = "cacheFields element not found in pivot cache definition."
        raise PivotCacheError(msg)

    existing_fields = [
        field.get("name")
        for field in cache_fields.findall(f"{{{ns}}}cacheField")
        if field.get("name")
    ]

    missing = [name for name in table_columns if name not in existing_fields]
    if not missing:
        return False

    for name in missing:
        new_field = etree.SubElement(cache_fields, f"{{{ns}}}cacheField")
        new_field.set("name", name)
        shared_items = etree.SubElement(new_field, f"{{{ns}}}sharedItems")
        shared_items.set("count", "0")

    cache_fields.set(
        "count",
        str(len(cache_fields.findall(f"{{{ns}}}cacheField"))),
    )
    return True


def _get_main_namespace(tree: etree._ElementTree) -> str:
    root = tree.getroot()
    ns = root.nsmap.get(None) or root.nsmap.get("main")
    return ns or _NS_MAIN



"""Custom exception hierarchy for pivoteer."""


class PivoteerError(Exception):
    """Base exception for pivoteer errors."""


class TemplateNotFoundError(PivoteerError):
    """Raised when the input template cannot be located or opened."""


class TableNotFoundError(PivoteerError):
    """Raised when the target Excel table cannot be resolved."""


class XmlStructureError(PivoteerError):
    """Raised when expected XML structures are missing or malformed."""


class InvalidDataError(PivoteerError):
    """Raised when the supplied data cannot be injected."""


class WriteError(PivoteerError):
    """Raised when writing the output file fails."""


class PivotCacheError(PivoteerError):
    """Raised when pivot cache metadata cannot be updated."""



"""High-level orchestration for XML updates."""

from __future__ import annotations

import logging
import zipfile
from pathlib import Path

import pandas as pd
from lxml import etree

from pivoteer.exceptions import InvalidDataError, TableNotFoundError
from pivoteer.models import TableRef, WorkbookMap
from pivoteer.pivot_cache_updater import sync_cache_fields
from pivoteer.table_resizer import TableResizer
from pivoteer.utils import parse_a1_range
from pivoteer.xml_engine import XmlEngine

LOGGER = logging.getLogger(__name__)


class TemplateEngine:
    """Coordinates XmlEngine and TableResizer for template updates."""

    def __init__(self, template_path: Path) -> None:
        self._xml_engine = XmlEngine(template_path)
        self._table_resizer = TableResizer()
        self._workbook_map: WorkbookMap = self._xml_engine.build_workbook_map()
        self._tables: dict[str, TableRef] = dict(self._workbook_map.tables)
        self._modified_trees: dict[str, etree._ElementTree] = {}
        self._updated_tables: set[str] = set()

    @property
    def template_path(self) -> Path:
        return self._xml_engine.template_path

    def apply_dataframe(self, table_name: str, df: pd.DataFrame) -> None:
        """Inject a DataFrame into the target table and resize it."""
        table_ref = self._tables.get(table_name)
        if not table_ref:
            raise TableNotFoundError(f"Table not found: {table_name}")
        if df.empty:
            raise InvalidDataError(
                f"Table '{table_name}' requires data, but DataFrame was empty."
            )
        if df.columns.empty:
            raise InvalidDataError(
                f"Table '{table_name}' requires columns, but DataFrame has none."
            )

        data_rows = df.itertuples(index=False, name=None)
        rows = [list(row) for row in data_rows]
        row_count = len(rows)
        col_count = len(df.columns)

        (start_row, start_col), _ = parse_a1_range(table_ref.ref)
        data_start_row = start_row + 1

        with zipfile.ZipFile(self.template_path, "r") as archive:
            sheet_tree = self._read_xml_part(archive, table_ref.worksheet_path)
            self._xml_engine.inject_rows_inline_strings(
                sheet_tree, data_start_row, start_col, rows
            )
            self._modified_trees[table_ref.worksheet_path] = sheet_tree

            table_tree = self._read_xml_part(archive, table_ref.table_path)
            resize_result = self._table_resizer.resize_table(
                table_tree, data_rows=row_count, data_cols=col_count
            )
            self._modified_trees[table_ref.table_path] = table_tree

        self._tables[table_name] = TableRef(
            name=table_ref.name,
            sheet_name=table_ref.sheet_name,
            table_path=table_ref.table_path,
            worksheet_path=table_ref.worksheet_path,
            ref=resize_result.updated_ref,
        )
        self._updated_tables.add(table_name)

    def ensure_pivot_refresh_on_load(self) -> None:
        """Set refreshOnLoad=1 for all pivot cache definitions."""
        pivot_paths = self._workbook_map.pivot_cache_definition_paths.values()
        if not pivot_paths:
            return

        with zipfile.ZipFile(self.template_path, "r") as archive:
            for path in pivot_paths:
                tree = self._read_xml_part(archive, path)
                root = tree.getroot()
                root.set("refreshOnLoad", "1")
                self._modified_trees[path] = tree

    def sync_pivot_cache_fields(self) -> None:
        """Append missing pivot cache fields for updated tables."""
        if not self._updated_tables:
            return

        for table_name in sorted(self._updated_tables):
            updated_parts = sync_cache_fields(self._workbook_map, table_name)
            for path, tree in updated_parts.items():
                self._modified_trees[path] = tree

    def get_modified_parts(self) -> dict[str, bytes]:
        """Serialize modified XML trees to bytes for writing."""
        parts: dict[str, bytes] = {}
        for path, tree in self._modified_trees.items():
            parts[path] = etree.tostring(
                tree, encoding="UTF-8", xml_declaration=True, standalone="yes"
            )
        return parts

    def _read_xml_part(self, archive: zipfile.ZipFile, path: str) -> etree._ElementTree:
        cached = self._modified_trees.get(path)
        if cached is not None:
            return cached
        return self._xml_engine.read_xml(archive, path)



"""Unit test for pivot cache refresh flag logic."""

from __future__ import annotations

from lxml import etree


def test_pivot_cache_refresh_flag() -> None:
    xml = (
        "<pivotCacheDefinition "
        'xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main" />'
    )
    tree = etree.fromstring(xml.encode("utf-8")).getroottree()
    root = tree.getroot()

    root.set("refreshOnLoad", "1")

    assert root.get("refreshOnLoad") == "1"



"""Integration test for real pivot cache preservation."""

from __future__ import annotations

import zipfile
from pathlib import Path

import pandas as pd
import pytest
from lxml import etree

from pivoteer.core import Pivoteer


def test_real_pivot_refresh_flag(tmp_path: Path) -> None:
    fixture_path = Path("tests/fixtures/real_pivot.xlsx")
    if not fixture_path.exists():
        pytest.skip("Fixture not found: tests/fixtures/real_pivot.xlsx")

    df = pd.DataFrame(
        {
            "Spalte1": ["A", "B", "C", "D", "E"],
            "Spalte2": [1, 2, 3, 4, 5],
        }
    )

    output_path = tmp_path / "real_pivot_output.xlsx"
    pivoteer = Pivoteer(fixture_path)
    pivoteer.apply_dataframe("TestTable", df)
    pivoteer.save(output_path)

    pivot_definition_path = None
    refresh_flag = None
    with zipfile.ZipFile(output_path, "r") as archive:
        for filename in archive.namelist():
            if "pivotCacheDefinition" in filename and filename.endswith(".xml"):
                pivot_definition_path = filename
                tree = etree.fromstring(archive.read(filename)).getroottree()
                root = tree.getroot()
                refresh_flag = root.get("refreshOnLoad")
                break

    print(f"Pivot cache path: {pivot_definition_path}")
    print(f"refreshOnLoad: {refresh_flag}")

    assert pivot_definition_path is not None, "No pivot cache definition found."
    assert refresh_flag == "1", "Pivot cache refreshOnLoad was not set to '1'."



"""Low-level XML and ZIP handling for pivoteer."""

from __future__ import annotations

import logging
import posixpath
import zipfile
from pathlib import Path

import pandas as pd
from lxml import etree

from pivoteer.exceptions import (
    InvalidDataError,
    TemplateNotFoundError,
    XmlStructureError,
)
from pivoteer.models import TableRef, WorkbookMap, WorksheetInfo
from pivoteer.utils import build_a1_cell

LOGGER = logging.getLogger(__name__)

_NS_MAIN = "http://schemas.openxmlformats.org/spreadsheetml/2006/main"
_NS_REL = "http://schemas.openxmlformats.org/officeDocument/2006/relationships"
_NS_PKG_REL = "http://schemas.openxmlformats.org/package/2006/relationships"

_NSMAP_MAIN = {"main": _NS_MAIN}
_NSMAP_REL = {"rel": _NS_REL}
_NSMAP_PKG = {"rel": _NS_PKG_REL}


def read_xml_part(archive: zipfile.ZipFile, path: str) -> etree._ElementTree:
    """Read an XML part from a ZIP archive and return an lxml ElementTree."""
    try:
        data = archive.read(path)
    except KeyError as exc:
        raise XmlStructureError(f"Missing XML part: {path}") from exc
    parser = etree.XMLParser(remove_blank_text=False)
    return etree.fromstring(data, parser).getroottree()


class XmlEngine:
    """Provides ZIP IO and XML manipulation for Excel workbooks."""

    def __init__(self, template_path: Path) -> None:
        if not template_path.exists():
            raise TemplateNotFoundError(f"Template not found: {template_path}")
        self._template_path = template_path

    @property
    def template_path(self) -> Path:
        return self._template_path

    def build_workbook_map(self) -> WorkbookMap:
        """Build a map of worksheets, tables, and pivot caches."""
        with zipfile.ZipFile(self._template_path, "r") as archive:
            workbook_tree = self._read_xml(archive, "xl/workbook.xml")
            rels_tree = self._read_xml(archive, "xl/_rels/workbook.xml.rels")

            worksheets = self._parse_worksheets(workbook_tree, rels_tree)
            tables = self._parse_tables(archive, worksheets)
            pivot_cache_paths = self._parse_pivot_caches(rels_tree)

        return WorkbookMap(
            template_path=self._template_path,
            worksheets=worksheets,
            tables=tables,
            pivot_cache_definition_paths=pivot_cache_paths,
        )

    def read_sheet_xml(
        self, archive: zipfile.ZipFile, worksheet_path: str
    ) -> etree._ElementTree:
        """Read worksheet XML as an lxml tree."""
        return self._read_xml(archive, worksheet_path)

    def write_sheet_xml(
        self, archive: zipfile.ZipFile, worksheet_path: str, tree: etree._ElementTree
    ) -> None:
        """Write worksheet XML back into the archive."""
        xml_bytes = etree.tostring(
            tree, encoding="UTF-8", xml_declaration=True, standalone="yes"
        )
        self._write_xml(archive, worksheet_path, xml_bytes)

    def inject_rows_inline_strings(
        self,
        tree: etree._ElementTree,
        start_row: int,
        start_col: int,
        rows: list[list[object]],
    ) -> None:
        """Inject data rows into sheetData using inline strings for text."""
        if start_row < 1 or start_col < 1:
            raise InvalidDataError("Start row/col must be >= 1.")
        if not rows:
            LOGGER.warning("No rows provided for injection; worksheet left unchanged.")
            return

        sheet_data = tree.find(".//main:sheetData", namespaces=_NSMAP_MAIN)
        if sheet_data is None:
            raise XmlStructureError("sheetData element not found.")

        for row_offset, row_values in enumerate(rows):
            row_idx = start_row + row_offset
            row_element = self._find_or_create_row(sheet_data, row_idx)

            for col_offset, value in enumerate(row_values):
                col_idx = start_col + col_offset
                cell_ref = build_a1_cell(row_idx, col_idx)
                cell = self._find_or_create_cell(row_element, cell_ref)
                self._set_cell_value_inline(cell, value)

        self._sort_rows(sheet_data)

    def _parse_worksheets(
        self,
        workbook_tree: etree._ElementTree,
        rels_tree: etree._ElementTree,
    ) -> dict[str, WorksheetInfo]:
        sheet_nodes = workbook_tree.findall(".//main:sheets/main:sheet", _NSMAP_MAIN)
        rel_map = self._parse_relationships(rels_tree)

        worksheets: dict[str, WorksheetInfo] = {}
        for sheet in sheet_nodes:
            name = sheet.get("name")
            sheet_id = sheet.get("sheetId")
            rel_id = sheet.get(f"{{{_NS_REL}}}id")
            if not name or not sheet_id or not rel_id:
                raise XmlStructureError("Worksheet metadata is incomplete.")

            target = rel_map.get(rel_id)
            if not target:
                raise XmlStructureError(f"Missing relationship for sheet {name}.")
            path = f"xl/{target}"

            worksheets[name] = WorksheetInfo(
                name=name, sheet_id=sheet_id, path=path, rel_id=rel_id
            )

        return worksheets

    def _parse_tables(
        self,
        archive: zipfile.ZipFile,
        worksheets: dict[str, WorksheetInfo],
    ) -> dict[str, TableRef]:
        tables: dict[str, TableRef] = {}
        for worksheet in worksheets.values():
            sheet_tree = self._read_xml(archive, worksheet.path)
            table_parts = sheet_tree.findall(
                ".//main:tableParts/main:tablePart", _NSMAP_MAIN
            )
            if not table_parts:
                continue

            rels_path = self._sheet_rels_path(worksheet.path)
            rels_tree = self._read_xml(archive, rels_path)
            rel_map = self._parse_relationships(rels_tree)

            for table_part in table_parts:
                rel_id = table_part.get(f"{{{_NS_REL}}}id")
                if not rel_id:
                    continue
                target = rel_map.get(rel_id)
                if not target:
                    continue
                table_path = self._normalize_rel_target(worksheet.path, target)
                table_tree = self._read_xml(archive, table_path)
                table_node = table_tree.getroot()
                name = table_node.get("name")
                ref = table_node.get("ref")
                if not name or not ref:
                    raise XmlStructureError("Table definition missing name or ref.")

                tables[name] = TableRef(
                    name=name,
                    sheet_name=worksheet.name,
                    table_path=table_path,
                    worksheet_path=worksheet.path,
                    ref=ref,
                )

        return tables

    def _parse_pivot_caches(self, rels_tree: etree._ElementTree) -> dict[str, str]:
        rel_map = self._parse_relationships(rels_tree)
        cache_paths: dict[str, str] = {}
        for rel_id, target in rel_map.items():
            if "pivotCache" in target:
                cache_paths[rel_id] = f"xl/{target}"
        return cache_paths

    def read_xml(self, archive: zipfile.ZipFile, path: str) -> etree._ElementTree:
        return read_xml_part(archive, path)

    # Backward-compatible alias
    _read_xml = read_xml

    def _write_xml(self, archive: zipfile.ZipFile, path: str, data: bytes) -> None:
        archive.writestr(path, data)

    def _parse_relationships(self, rels_tree: etree._ElementTree) -> dict[str, str]:
        rels: dict[str, str] = {}
        rel_nodes = rels_tree.findall(".//rel:Relationship", _NSMAP_PKG)
        for rel in rel_nodes:
            rel_id = rel.get("Id")
            target = rel.get("Target")
            if rel_id and target:
                rels[rel_id] = target
        return rels

    def _sheet_rels_path(self, worksheet_path: str) -> str:
        filename = Path(worksheet_path).name
        parent = Path(worksheet_path).parent
        return str(parent / "_rels" / f"{filename}.rels")

    def _normalize_rel_target(self, worksheet_path: str, target: str) -> str:
        base_dir = posixpath.dirname(worksheet_path)
        normalized = posixpath.normpath(posixpath.join(base_dir, target))
        if not normalized.startswith("xl/"):
            normalized = f"xl/{normalized.lstrip('./')}"
        return normalized

    def _find_or_create_row(
        self, sheet_data: etree._Element, row_idx: int
    ) -> etree._Element:
        row = sheet_data.find(f"main:row[@r='{row_idx}']", namespaces=_NSMAP_MAIN)
        if row is not None:
            return row
        row = etree.SubElement(sheet_data, f"{{{_NS_MAIN}}}row")
        row.set("r", str(row_idx))
        return row

    def _find_or_create_cell(
        self, row: etree._Element, cell_ref: str
    ) -> etree._Element:
        cell = row.find(f"main:c[@r='{cell_ref}']", namespaces=_NSMAP_MAIN)
        if cell is not None:
            return cell
        cell = etree.SubElement(row, f"{{{_NS_MAIN}}}c")
        cell.set("r", cell_ref)
        return cell

    def _set_cell_value_inline(self, cell: etree._Element, value: object) -> None:
        for child in list(cell):
            cell.remove(child)

        if value is None or self._is_missing(value):
            cell.attrib.pop("t", None)
            cell.attrib.pop("s", None)
            return

        if isinstance(value, (int, float)):
            cell.attrib.pop("t", None)
            v = etree.SubElement(cell, f"{{{_NS_MAIN}}}v")
            v.text = str(value)
            return

        text_value = value.isoformat() if hasattr(value, "isoformat") else str(value)

        cell.set("t", "inlineStr")
        inline = etree.SubElement(cell, f"{{{_NS_MAIN}}}is")
        text = etree.SubElement(inline, f"{{{_NS_MAIN}}}t")
        text.text = text_value

    def _is_missing(self, value: object) -> bool:
        try:
            return bool(pd.isna(value))
        except (TypeError, ValueError):
            return False

    def _sort_rows(self, sheet_data: etree._Element) -> None:
        rows = sheet_data.findall("main:row", namespaces=_NSMAP_MAIN)
        rows_sorted = sorted(rows, key=lambda elem: int(elem.get("r", "0")))
        for row in rows:
            sheet_data.remove(row)
        for row in rows_sorted:
            sheet_data.append(row)



"""Table resizing logic for Excel ListObjects."""

from __future__ import annotations

from dataclasses import dataclass

from lxml import etree

from pivoteer.exceptions import XmlStructureError
from pivoteer.utils import build_a1_range, parse_a1_range


@dataclass(frozen=True)
class TableResizeResult:
    """Result of a table resize operation."""

    original_ref: str
    updated_ref: str


class TableResizer:
    """Calculates and applies table range updates."""

    def resize_table(
        self,
        table_tree: etree._ElementTree,
        data_rows: int,
        data_cols: int,
    ) -> TableResizeResult:
        if data_rows < 0 or data_cols < 1:
            raise ValueError("Data rows must be >= 0 and data cols >= 1.")

        table = table_tree.getroot()
        ref = table.get("ref")
        if not ref:
            raise XmlStructureError("Table ref attribute missing.")

        (start_row, start_col), _ = parse_a1_range(ref)
        header_rows = 1
        end_row = start_row + header_rows + data_rows - 1
        end_col = start_col + data_cols - 1

        updated_ref = build_a1_range(start_row, start_col, end_row, end_col)
        table.set("ref", updated_ref)
        return TableResizeResult(original_ref=ref, updated_ref=updated_ref)



"""Data models for workbook metadata and table references."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path


@dataclass(frozen=True)
class TableRef:
    """Represents a table (ListObject) and its XML references."""

    name: str
    sheet_name: str
    table_path: str
    worksheet_path: str
    ref: str


@dataclass(frozen=True)
class WorksheetInfo:
    """Represents worksheet metadata from workbook relationships."""

    name: str
    sheet_id: str
    path: str
    rel_id: str


@dataclass(frozen=True)
class WorkbookMap:
    """Holds resolved workbook references used for XML updates."""

    template_path: Path
    worksheets: dict[str, WorksheetInfo]
    tables: dict[str, TableRef]
    pivot_cache_definition_paths: dict[str, str]
    shared_strings_path: str | None = None
